#
# rubyrefmparser.rb
#
#   Copyright (c) 2001 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

module ReFe

  class RubyReferenceManualParser

    ::ReFe.declear_parser self

    def self.process( files, db, msgout )
      db.m_doc_updating {
          files.each do |fname|
            msgout.print_filename fname
            File.open( fname ) {|f|
                new.process f, db
            }
          end
      }
    end


    def process( f, db )
      parse(f).each do |c,methods|
        if methods.empty? then
          db.register_class c.chop
        else
          methods.each do |m,str|
            db.register_mdoc c+m, str
          end
        end
      end
    end


    const = '[A-Z][a-zA-Z0-9_]+'

    CLASS_BEGIN = {
      /\A= (#{const}(?:::#{const})*)\s*\z/n => false,
      /\A= ȹߴؿ\Z/e                   => 'Kernel',
      /\A=+ ¤Υ饹Υ饹᥽å/e  => 'Struct::XXXX',
    }

    CLASSMETHOD_BEGIN = [
      /\A=+ 饹᥽å/e,
      /\A=+ ⥸塼ؿ/e,
      /\A=+ ⥸塼°/e,
      /\A=+ ⥸塼᥽å/e
    ]

    METHOD_BEGIN = [
      /\A=+ ᥽å/e,
      /\A=+ ץ饤١ȥ᥽å/e
    ]

    def parse( f )
      table = {}
      mt = st = nil
      klass = buf = s = nil
      isalias = false

      f.each do |line|
        case line
        when /\A=+ /
          isalias = false
          untag! buf
          buf = nil

          case line
          when * CLASSMETHOD_BEGIN
            s = true
          when * METHOD_BEGIN
            s = false
          else
            if /\A= / === line then
              klass = nil
            end
            CLASS_BEGIN.each do |re,name|
              if m = re.match( line ) then
                klass = name ? name : m[1]
                mt = (table[klass + '#'] ||= {})
                st = (table[klass + '.'] ||= {})
                break
              end
            end
          end

        when /\A: /
          if klass and n = getmname(line) then
            buf = register_method( n, line, buf, isalias, (s ? st : mt) )
            isalias = true
          else
            untag! buf
            buf = nil
            isalias = false
          end
        
        else
          isalias = false
          buf << line if buf
        end
      end

      table
    end

    def register_method( method, line, buf, isalias, table )
      if t = table[method] then
        t << line
        buf = t
      elsif isalias then
        buf << line
        table[method] = buf
      else
        untag! buf
        buf = line
        table[method] = buf
      end
      buf
    end

    METHOD_RE = [
      [/\A:\s*(\w+[?!]?)\s*(?:[\(\{]|\z)/    , ''   ],# name(arg), name{}, name
      [/\A:\s*(\w+)\s*\=\s*\w+/              , '='  ],# name=
      [/\A:\s*\w+\[.*\]=/                    , '[]='],# self[key]=
      [/\A:\s*\w+\[.*\]/                     , '[]' ],# self[key]
      [/\A:\s*self\s*([+-<>=~*^&|%]+)\s*\w+/ , ''   ],# self + other
      [/\A:\s*([+\-~`])\s*\w+/               , ''   ] # ~ self
    ]

    def getmname( line )
      i = _getmname(line)
      printf "%15s %s\n", i.inspect, line.strip if $DEBUG and not i
      i
    end

    def _getmname( line )
      METHOD_RE.each do |re,s|
        if m = re.match(line) then
          return case s.size
                 when 0 then m[1]
                 when 1 then m[1] + s
                 else        s
                 end
        end
      end
      nil
    end

    def untag!( str )
      return unless str
      str.gsub!( /\(\(-.*-\)\)/, '' )
      str.gsub!( /\(\([{|'<]|[}|'>]\)\)/, '' )
    end
  
  end

end
